(* file: assoc_list.sml *)
(* description: functions for association list *)
(* author: Takahito Aoto *)

signature ASSOC_LIST  =
sig
    val find: ''a -> (''a * 'b) list -> 'b option
    val find': ('a * 'a -> bool) -> 'a -> ('a * 'b) list -> 'b option
    val singleton:  (''a * 'b) -> (''a * 'b) list
    val empty: (''a * 'b) list

    val add: (''a * ''b) -> (''a * ''b) list -> (''a * ''b) list option
    val add': (('a * 'a -> bool) * ('b * 'b -> bool)) 
	      -> ('a * 'b) -> ('a * 'b) list -> ('a * 'b) list option
end

structure AssocList : ASSOC_LIST=
struct 

fun find x [] = NONE
  | find x ((k,v)::ys) = if x = k then SOME v else find x ys
fun find' pred x [] = NONE
  | find' pred x ((k,v)::ys) = if pred (x,k) then SOME v else find' pred x ys

val empty = []
fun singleton (k,v) = ((k,v)::[])

fun add (k,v) ys = case find k ys of
		       SOME w => if v = w then SOME ys
				 else NONE
		     | NONE => SOME ((k,v)::ys)

fun add' (predKey,predVal) (k,v) ys
    = case find' predKey k ys of
	  SOME w => if predVal (v, w) then SOME ys
		    else NONE 
	| NONE => SOME ((k,v)::ys)

end
