(* file: atom.sml *)
(* description: atoms and permutation for nominal terms *)
(* author: Takahito Aoto *)

signature ATOM = 
sig 
    eqtype atom
    type key = atom
    val toString: key -> string
    val fromString: string -> key
    val compare: atom * atom -> order
    val swap: atom * atom -> atom -> atom
    val newAtoms: key list -> int -> key list
    val newAtom: key list -> key
end

structure Atom : ATOM =
struct 

local 
    structure L = List
    structure LP = ListPair
    structure SU = StringUtil
    structure TU = TestUtil
in

datatype atom = Atom of string
type key = atom

fun toString (Atom name) = name
fun fromString str = Atom str

fun compare (a,b) = String.compare (toString a, toString b)
val _ = TU.test "compare" 
	     [(compare (fromString "a", fromString "b") = LESS),
	      (compare (fromString "a", fromString "a") = EQUAL),
	      (compare (fromString "a1", fromString "a") = GREATER)]

fun swap (b,c) a = if a = b then c else if a = c then b else a
val _ = TU.test "swap" 
	     [(swap (fromString "a", fromString "b") (fromString "b") = (fromString "a")),
	      (swap (fromString "a", fromString "b") (fromString "a") = (fromString "b")),
	      (swap (fromString "a", fromString "b") (fromString "c") = (fromString "c"))]

(* creating fresh atoms *)
val defaultNames = ["a", "b", "c", "d", "e"]
fun newAtoms used num = L.map (fn str => Atom str) (SU.newNames defaultNames (L.map toString used) num)
val _ = TU.test "newAtoms" 
	     [(newAtoms [fromString "a", fromString "b"] 5 = 
			       [fromString "c", fromString "d", fromString "e", fromString "a0", fromString "b0"])]

fun newAtom used = hd (newAtoms used 1)
val _ = TU.test "newAtom" 
	     [ (newAtom [fromString "a", fromString "b"] = (fromString "c")),
	       (newAtom [fromString "a", fromString "b", fromString "c", fromString "d", fromString "e", 
			 fromString "a0"] = (fromString "b0")),
	       (newAtom [fromString "a", fromString "b", fromString "c", fromString "d", fromString "e", 
			 fromString "a0", fromString "b0", fromString "c0", fromString "d0", fromString "e0", 
			 fromString "a1"] = (fromString "b1")),
	       (newAtom [fromString "a", fromString "b", fromString "c", fromString "d", fromString "e", 
				       fromString "a0", fromString "b1", fromString "c0", fromString "d0", fromString "e0"]
				       = (fromString "b0"))]

end (* of local *)
end
