(* file: equiv.sml *)
(* description: equivalence of nominal terms *)
(* author: Takahito Aoto *)

signature EQUIV = 
sig 
    type var = Var.key
    type atom = Atom.key
    type perm = Perm.key
    type fun_key = Fun.key
    type term = Term.key
    type fresh_env = FreshEnv.key

    val aeq: fresh_env -> term * term -> bool
end

structure Equiv : EQUIV =
struct
local 
    structure B = Bool
    structure L = List
    structure LP = ListPair
    structure LU = ListUtil
    structure PU = PrintUtil
    structure SU = StringUtil
    structure T = Term
    structure TU = TestUtil
in

type var = Var.key
type atom = Atom.key
type perm = Perm.key
type fun_key = Fun.key
type term = Term.key
type fresh_env = FreshEnv.key

(* env |- s =_\alpha t *)
fun aeq env (T.Atom a, T.Atom b) = (a = b)
  | aeq env (T.Susp (pi1,X1), T.Susp (pi2,X2)) = (X1 = X2) 
					   andalso L.all (fn a => LU.member (a,X1) env) (Perm.disagreements (pi1,pi2))
(***  
   | aeq env (T.Abst (a,t), T.Abst (b,s)) = if a = b 
				       then aeq env (t,s)
				       else (aeq env (T.actDot (Perm.single (a,b)) t, s) andalso FreshEnv.isFresh env b t)
***) (*** changed the employing definition !! 2015/12/29 ***)
  | aeq env (T.Abst (a,t), T.Abst (b,s)) = if a = b 
					   then aeq env (t,s)
					   else aeq env (t, T.actDot (Perm.single (a,b)) s) andalso FreshEnv.isFresh env a s
  | aeq env (T.Fun (f,t), T.Fun (g,s)) = f = g andalso aeq env (t,s)
  | aeq env (T.Tuple ts, T.Tuple ss) = (L.length ts) = (L.length ss) 
				   andalso LP.all (aeq env) (ts,ss)
  | aeq env _ = false

val _ = TU.test "aeq"
		[ (aeq [] (T.fromString "a", T.fromString "a") = true),
		  (aeq [] (T.fromString "a", T.fromString "b") = false),
		  (aeq [] (T.fromString "[(b a)]X", T.fromString "[(a b)]X") = true),
		  (aeq [] (T.fromString "[a][(b a)]X", T.fromString "[b]X") = false),
		  (aeq [] (T.fromString "[(b a)(b a)]X", T.fromString "[]X") = true),
		  (aeq (FreshEnv.fromString "a#X") (T.fromString "[a][(b a)]X", T.fromString "[b][]X") = true),
		  (aeq (FreshEnv.fromString "b#X") (T.fromString "[a][]X", T.fromString "[b][(b a)]X") = true),
		  (aeq (FreshEnv.fromString "a#X,b#X") (T.fromString "[b][]X", T.fromString "[b][(b a)]X") = true),
		  (aeq [] (T.fromString "(app <a,b>)", T.fromString "(app <a,b>)") = true),
		  (aeq [] (T.fromString "[c](app <a,c>)", T.fromString "[b](app <a,b>)") = true),
		  (aeq [] (T.fromString "[c](app <a,b>)", T.fromString "[b](app <a,c>)") = false),
		  (aeq [] (T.fromString "(sub <[c](app <a,c>),b>)", T.fromString "(sub <[b](app <a,b>),b>)") = true),
  	          (aeq [] (T.fromString "(lam [c](sub <[a](app <c,a>),b>))", T.fromString "(lam [a](sub <[b](app <a,b>),b>))") = true),
		  (aeq [] (T.fromString "(forall [b](forall [a][(a b)]X))", T.fromString "(forall [a](forall [b][]X))") = true)]

end (* of local *)		     

end
