(* file: io_tpdb.yacc *)
(* description: reading an input file in TPDB format *)
(* author: Takahito Aoto *)

signature IO_TPDB = sig
    type var = Var.key
    type atom = Atom.key
    type perm = Perm.key
    type fun_key = Fun.key
    type term = Term.key
    type fresh_env = FreshEnv.key
    type rule = Rule.rule
    type rules = Rule.rules
    val rdFile : string -> rules
end

structure IOTpdb : IO_TPDB = struct
type var = Var.key
type atom = Atom.key
type perm = Perm.key
type fun_key = Fun.key
type term = Term.key
type fresh_env = FreshEnv.key
type rule = Rule.rule
type rules = Rule.rules
structure PU = PrintUtil

exception FormatError

structure TpdbLrVals = TpdbLrValsFun (structure Token = LrParser.Token);
structure TpdbLex = TpdbLexFun (structure Tokens = TpdbLrVals.Tokens);
structure TpdbParser = Join (structure ParserData = TpdbLrVals.ParserData;
                             structure Lex = TpdbLex;
                             structure LrParser = LrParser);
fun parserStream istream = 
    let val lexer = TpdbParser.makeLexer 
                        (fn n=>TextIO.inputN (istream, n));
        val print_error = (fn (s, i, _) =>
                              (print ("lex: " ^ s ^ ", line " ^ Int.toString(i) ^ "\n");
                               raise FormatError))
    in #1 (TpdbParser.parse (0, lexer, print_error, ()))
    end;

fun rdFile input = 
    let (* val ins = TextIO.openIn filename *)
	val ins = TextIO.openString input
        val rules = parserStream ins
    in rules
    end

end
