(* file: lexical.sml *)
(* description: lexer generator *)
(* author: Takahito Aoto *)

signature LEXICAL =
  sig
  datatype token = Id of string | Key of string
  val scan : string -> token list
  end;

signature KEYWORD =
  sig
  val words : string list
  and symbols : char list
  end;

functor Lexical (Keyword: KEYWORD) : LEXICAL =
  struct
  datatype token = Key of string | Id of string;

  fun reserved c = ListUtil.member c Keyword.symbols

  fun scanning (toks, ss) =
      case Substring.getc ss of
	  NONE => rev toks
	| SOME (c, ss1) => 
	  if not (Char.isSpace c)
	  then let val (id, ss2) = Substring.splitl (fn c => not (Char.isSpace c)) ss
		   val str = Substring.string id
	       in if ListUtil.member str Keyword.words
		  then scanning ((Key str)::toks, ss2)
		  else if reserved c
		  then scanning ((Key (String.str c))::toks, ss1)
		  else let val (id, ss3) = Substring.splitl (fn c => not (Char.isSpace c) andalso (not (reserved c))) ss
			   val str = Substring.string id
		       in scanning ((Id str)::toks, ss3)
		       end
	       end
	  else
	      scanning (toks, Substring.dropl Char.isSpace ss);

  fun scan a = scanning ([], Substring.full a);
  end;



