(* file: list_util.sml *)
(* description: utility functions for list *)
(* author: Takahito Aoto *)

signature LIST_UTIL = 
sig 
    val member: ''a  -> ''a list -> bool
    val member': ('a * 'a -> bool) -> 'a  -> 'a list -> bool
    val add: ''a  -> ''a list -> ''a list
    val add':  ('a * 'a -> bool) -> 'a  -> 'a list -> 'a list
    val delete: ''a  -> ''a list -> ''a list
    val delete': ('a * 'a -> bool) -> 'a  -> 'a list -> 'a list
    val toSet: ''a list -> ''a list
    val toSet': ('a * 'a -> bool) -> 'a list -> 'a list
    val isDistinct: ''a list -> bool
    val isDistinct': ('a * 'a -> bool) -> 'a list -> bool
    val isSubset: ''a list  -> ''a list -> bool
    val isSubset': ('a * 'a -> bool) -> 'a list  -> 'a list -> bool
    val isEqset: ''a list  ->  ''a list -> bool
    val isEqset': ('a * 'a -> bool) -> 'a list -> 'a list -> bool
    val isDisjoint: ''a list * ''a list -> bool
    val isDisjoint': ('a * 'a -> bool) -> 'a list * 'a list -> bool
    val differenceByOne: ''a list * ''a list -> ''a list 
    val differenceByOne': ('a * 'a -> bool) -> 'a list * 'a list -> 'a list 
    val differenceByAll: ''a list * ''a list -> ''a list 
    val differenceByAll': ('a * 'a -> bool) -> 'a list * 'a list -> 'a list 
    val union: ''a list * ''a list -> ''a list
    val union': ('a * 'a -> bool) -> 'a list * 'a list -> 'a list
    val unionAll: (''a list) list -> ''a list
    val unionAll': ('a * 'a -> bool) -> ('a list) list -> 'a list

    val mapAppend: ('a -> 'b list) -> 'a list -> 'b list
    val mapUnion: ('a -> ''b list) -> 'a list -> ''b list
    val mapUnion': ('b * 'b -> bool) -> ('a -> 'b list) -> 'a list -> 'b list
    val replaceNth: ('a list * int * 'a) -> 'a list
    val takeUntilWithRest : ('a -> bool) -> 'a list -> 'a list * 'a list

    val toString: ('a -> string) -> 'a list -> string
    val toStringComma: ('a -> string) -> 'a list -> string
    val toStringSpace: ('a -> string) -> 'a list -> string
    val toStringAst: ('a -> string) -> 'a list -> string
    val toStringPlus: ('a -> string) -> 'a list -> string
    val toStringPlusRound: ('a -> string) -> 'a list -> string
    val toStringSlashSquare: ('a -> string) -> 'a list -> string

    val toStringCommaSquare: ('a -> string) -> 'a list -> string
    val toStringCommaAngle: ('a -> string) -> 'a list -> string
    val toStringCommaCurly: ('a -> string) -> 'a list -> string
    val toStringCommaRound: ('a -> string) -> 'a list -> string
    val toStringCommaLnSquare: ('a -> string) -> 'a list -> string
    val toStringCommaLnCurly: ('a -> string) -> 'a list -> string

    val toStringSpaceSquare: ('a -> string) -> 'a list -> string
    val toStringSpaceAngle: ('a -> string) -> 'a list -> string
    val toStringSpaceCurly: ('a -> string) -> 'a list -> string
    val toStringSpaceRound: ('a -> string) -> 'a list -> string
    val toStringSpaceLnSquare: ('a -> string) -> 'a list -> string
    val toStringSpaceLnCurly: ('a -> string) -> 'a list -> string

    val toStringSemicolonSquare: ('a -> string) -> 'a list -> string
    val toStringSemicolonAngle: ('a -> string) -> 'a list -> string
    val toStringSemicolonCurly: ('a -> string) -> 'a list -> string
    val toStringSemicolonRound: ('a -> string) -> 'a list -> string
    val toStringSemicolonLnSquare: ('a -> string) -> 'a list -> string
    val toStringSemicolonLnCurly: ('a -> string) -> 'a list -> string


end

structure ListUtil : LIST_UTIL =
struct 

local 
    structure L = List
in

fun member x ys = List.exists (fn y => x = y) ys
fun member' pred x ys = List.exists (fn y => pred (x,y)) ys

(* 集合としての演算 *)
fun add x ys = if member x ys then ys else x::ys
fun add' pred x ys = if member' pred x ys then ys else x::ys

fun delete x [] = []
  | delete x (y::ys) = if x = y then delete x ys else y::(delete x ys)
fun delete' pred x [] = []
  | delete' pred x (y::ys) = if pred (x,y) then delete' pred x ys else y::(delete' pred x ys)

(* corrected 2016/12/29 *)
fun toSet [] = []
  | toSet (x::xs) = let val xset = toSet xs 
		    in if member x xset then xset else x::xset
		    end
fun toSet' pred  [] = []
  | toSet' pred  (x::xs) = let val xset = toSet' pred xs 
			   in if member' pred x xset then xset else x::xset
			   end

fun isDistinct [] = true
  | isDistinct (x::xs) = not (member x xs) andalso isDistinct xs
fun isDistinct' pred [] = true
  | isDistinct' pred (x::xs) = not (member' pred x xs) andalso isDistinct' pred xs

fun isSubset xs ys = L.all (fn x => member x ys) xs
fun isSubset' pred xs ys = L.all (fn x => member' pred x ys) xs

fun isEqset xs ys = (isSubset xs ys) andalso (isSubset ys xs)
fun isEqset' pred xs ys = (isSubset' pred xs ys) andalso (isSubset' pred ys xs)

fun isDisjoint (xs,ys) = L.all (fn x => not (member x ys)) xs
			 andalso L.all (fn y => not (member y xs)) ys
fun isDisjoint' pred (xs,ys) = L.all (fn x => not (member' pred  x ys)) xs
			 andalso L.all (fn y => not (member' pred  y xs)) ys


fun deleteOne x ys = 
    let fun delete [] ans = rev ans
          | delete (y'::ys') ans = 
            if x = y' then L.revAppend (ans, ys')
            else delete ys' (y'::ans)
    in delete ys [] 
    end
fun deleteOne' pred x ys = 
    let fun delete [] ans = rev ans
          | delete (y'::ys') ans = 
            if pred (x,y') then L.revAppend (ans, ys')
            else delete ys' (y'::ans)
    in delete ys [] 
    end

fun deleteAll x ys = 
    let fun delete [] ans = rev ans
          | delete (y'::ys') ans = 
            if x = y' then delete ys' ans
            else delete ys' (y'::ans)
    in delete ys [] 
    end
fun deleteAll' pred  x ys = 
    let fun delete [] ans = rev ans
          | delete (y'::ys') ans = 
            if pred (x,y') then delete ys' ans
            else delete ys' (y'::ans)
    in delete ys [] 
    end

fun differenceByOne (xs,[]) = xs
  | differenceByOne (xs,y::ys) = differenceByOne (deleteOne y xs, ys)
fun differenceByOne' pred (xs,[]) = xs
  | differenceByOne' pred (xs,y::ys) = differenceByOne' pred (deleteOne' pred y xs, ys)

fun differenceByAll (xs,[]) = xs
  | differenceByAll (xs,y::ys) = differenceByAll (deleteAll y xs, ys)
fun differenceByAll' pred (xs,[]) = xs
  | differenceByAll' pred (xs,y::ys) = differenceByAll' pred (deleteAll' pred y xs, ys)

fun union ([],ys) = ys
  | union (x::xs, ys) = if member x ys 
			then union (xs,ys)
			else x :: union (xs,ys)
fun union' pred ([],ys) = ys
  | union' pred (x::xs, ys) = if member' pred x ys 
			      then union' pred (xs,ys)
			      else x :: union' pred (xs,ys)

fun unionAll xss = List.foldl union [] xss
fun unionAll' pred xss = List.foldl (union' pred) [] xss

fun mapAppend f [] =  []
  | mapAppend f (x::xs) =  (f x) @  (mapAppend f xs)

fun mapUnion f [] =  []
  | mapUnion f (x::xs) =  union (f x, mapUnion f xs)

fun mapUnion' pred f [] =  []
  | mapUnion' pred f (x::xs) =  union' pred (f x, mapUnion' pred f xs)

fun replaceNth (xs,i,y) = (L.take (xs,i)) @ [y] @ (L.drop (xs,i+1))

fun takeUntilWithRest pred [] = ([],[])
  | takeUntilWithRest pred (x::xs) = if pred x 
				     then ([x],xs)
				     else (fn (ys,zs) => (x::ys,zs)) (takeUntilWithRest pred xs)


fun toStringBuilt prElm (start,sep,stop) xs =
    let fun toStringSub [] = ""
	  | toStringSub [x] = (prElm x)
	  | toStringSub (x::xs)= (prElm x) ^ sep ^ (toStringSub xs)
    in  start ^ (toStringSub xs) ^ stop
    end

fun toString prElm xs = toStringBuilt prElm ("",  "", "") xs

fun toStringComma prElm xs = toStringBuilt prElm ("",  ",", "") xs
fun toStringSpace prElm xs = toStringBuilt prElm ("",  " ", "") xs
fun toStringAst prElm xs = toStringBuilt prElm   ("",  "*", "") xs
fun toStringPlus prElm xs = toStringBuilt prElm   ("",  "+", "") xs
fun toStringPlusRound prElm xs = toStringBuilt prElm   ("(",  "+", ")") xs
fun toStringSlashSquare prElm xs = toStringBuilt prElm   ("[",  "/", "]") xs

fun toStringSpaceSquare prElm xs = toStringBuilt prElm ("[",  " ",  "]") xs
fun toStringSpaceAngle prElm xs = toStringBuilt prElm  ("<",  " ",  ">") xs
fun toStringSpaceCurly prElm xs = toStringBuilt prElm  ("{",  " ",  "}") xs
fun toStringSpaceRound prElm xs = toStringBuilt prElm  ("(",  " ",  ")") xs
fun toStringSpaceLnSquare prElm xs = toStringBuilt prElm ("   [ ",  "\n     ",  " ]\n") xs
fun toStringSpaceLnCurly prElm xs = toStringBuilt prElm ("   { ",  "\n     ",  " }\n") xs


fun toStringCommaSquare prElm xs = toStringBuilt prElm ("[",  ",",  "]") xs
fun toStringCommaAngle prElm xs = toStringBuilt prElm  ("<",  ",",  ">") xs
fun toStringCommaCurly prElm xs = toStringBuilt prElm  ("{",  ",",  "}") xs
fun toStringCommaRound prElm xs = toStringBuilt prElm  ("(",  ",",  ")") xs
fun toStringCommaLnSquare prElm xs = toStringBuilt prElm ("   [ ",  ",\n     ",  " ]\n") xs
fun toStringCommaLnCurly prElm xs = toStringBuilt prElm ("   { ",  ",\n     ",  " }\n") xs

fun toStringSemicolonSquare prElm xs = toStringBuilt prElm ("[",  ";",  "]") xs
fun toStringSemicolonAngle prElm xs = toStringBuilt prElm  ("<",  ";",  ">") xs
fun toStringSemicolonCurly prElm xs = toStringBuilt prElm  ("{",  ";",  "}") xs
fun toStringSemicolonRound prElm xs = toStringBuilt prElm  ("(",  ";",  ")") xs
fun toStringSemicolonLnSquare prElm xs = toStringBuilt prElm ("   [ ",  ";\n     ",  " ]\n") xs
fun toStringSemicolonLnCurly prElm xs = toStringBuilt prElm ("   { ",  ";\n     ",  " }\n") xs


end (* of local *)

end
