(* file: tpdb.lex *)
(* description: lexer generator *)
(* author: Takahito Aoto *)

structure Tokens = Tokens;
type pos = int;
type svalue = Tokens.svalue;
type ('a,'b) token = ('a,'b) Tokens.token;
type lexresult = (svalue, pos) token;

open Tokens
val pos = ref 1;
val pr = fn s : string => TextIO.output(TextIO.stdErr, s)
val eof = fn () => (pos:=1; Tokens.EOF(!pos,!pos));
val text = ref (nil: string list)
val Add = fn s => (text := s::(!text))
%%
%header (functor TpdbLexFun (structure Tokens: Tpdb_TOKENS));
%s STRING;
idstr = [!|$%&'\*\+/;=?@\\^_`{}~a-zA-Z0-9\-\=\:\.];
%%
<INITIAL>\n	     => (pos:=(!pos)+1; lex()  (* Ԥ *));
<INITIAL>[\ \013
\t]+     => (lex() (* ϥå *));
<INITIAL>\->  	     => (Tokens.ARROW(!pos,!pos));
<INITIAL>\|-          => (Tokens.VDASH(!pos,!pos));
<INITIAL>\(    	     => (Tokens.LPAREN(!pos,!pos));
<INITIAL>\)    	     => (Tokens.RPAREN(!pos,!pos));
<INITIAL>\[    	     => (Tokens.LSQBRA(!pos,!pos));
<INITIAL>\]    	     => (Tokens.RSQBRA(!pos,!pos));
<INITIAL>\<    	     => (Tokens.LBRA(!pos,!pos));
<INITIAL>\>    	     => (Tokens.RBRA(!pos,!pos));
<INITIAL>\,    	     => (Tokens.COMMA(!pos,!pos));
<INITIAL>\#    	     => (Tokens.SHARP(!pos,!pos));
<INITIAL>"\""        => (text := nil; Add yytext;
                         YYBEGIN STRING; continue());
<INITIAL>VAR          => (Tokens.VAR(!pos,!pos));
<INITIAL>ATOMS        => (Tokens.ATOMS(!pos,!pos));
<INITIAL>RULES        => (Tokens.RULES(!pos,!pos));
<INITIAL>{idstr}+     => (Tokens.ID(yytext,!pos,!pos));
<INITIAL>. => (print ("lex omitted [" ^ yytext ^ "]\n");continue ());
<STRING>"\""	=> (Add yytext; YYBEGIN INITIAL; 
                    Tokens.STR(String.concat (List.rev (!text)),!pos,!pos));
<STRING>[^"\n]+ => (Add yytext; continue());
<STRING>\n	     => (Add yytext; pos:=(!pos)+1; continue());
