(* file: var.sml *)
(* description: variables for nominal terms *)
(* author: Takahito Aoto *)

signature VAR = 
sig
    eqtype key
    val name: key -> string
    val index: key -> int
    val incIndex: int -> key -> key
    val toString: key -> string
    val fromString: string -> key
    val newVars: key list -> int -> key list
end

structure Var : VAR =
struct

local
    structure L = List
    structure LU = ListUtil
    structure TU = TestUtil
in

datatype key = Var of string * int

fun name (Var (x,i)) = x
fun index (Var (x,i)) = i
fun incIndex n (Var (x,i)) = Var (x,i+n)

fun toString (Var (x,i)) =
    if i = 0  then x
    else if Int.> (i, 0) then x ^ "_" ^ (Int.toString i)
    else raise Fail "Error: Var.toString: var index out of range"

val _ = TestUtil.test "toString"
		[ (toString (Var ("a",1)) = "a_1"),
		  (toString (Var ("a",0)) = "a"),
		  (toString (Var ("aa",0)) = "aa")]

(* .+_[1-9][0-9]* となっていれば末尾を index とする *)
fun fromString str = 
    let val ss = Substring.full str
	val (body,index) = Substring.splitr Char.isDigit ss
	val isProperIndexing = Int.> (Substring.size body, 1)
			    andalso Substring.sub (body, Substring.size body - 1) = #"_"
			    andalso not (Substring.isEmpty index)
			    andalso Substring.sub (index, 0) <> #"0"
    in if isProperIndexing
       then Var (Substring.string (Substring.trimr 1 body),
		  valOf (Int.fromString (Substring.string index)))
       else Var (str,0)
    end

val _ = TestUtil.test "fromString" 
		      [(fromString "a_1" = Var ("a",1)),
		       (fromString "a_120" = Var ("a",120)),
		       (fromString "a12_34" = Var ("a12",34)),
		       (fromString "a12_034" = Var ("a12_034",0))]

(* creating fresh atoms *)
val defaultNames = ["X", "Y", "Z", "U", "V", "W"]
fun newVars used num = 
    let fun nvSub alle (ys,n) 
	    = let val freshes = LU.differenceByAll (alle, used)
	      in if null freshes
 		 then nvSub (L.map (incIndex 1) alle) (ys,n)
		 else if n <= (L.length freshes)
		 then ys @ (L.take (freshes,n))
		 else nvSub (L.map (incIndex 1) alle) (ys @ freshes, n - (L.length freshes))
	      end
    in nvSub (L.map fromString defaultNames) ([], num)
    end

val _ = TU.test "newVars" 
	     [ 
	       (newVars [fromString "X_1", fromString "Y", fromString "Z" ] 10 = 
		[fromString "X", fromString "U", fromString "V", fromString "W", fromString "Y_1",
		 fromString "Z_1", fromString "U_1", fromString "V_1", fromString "W_1", fromString "X_2"])
	     ]


end (* of local *)
end (* of struct *)
